{
  "type": "record",
  "name": "GlossaryNodeInfo",
  "namespace": "com.linkedin.pegasus2avro.glossary",
  "doc": "Properties associated with a GlossaryNode",
  "fields": [
    {
      "name": "definition",
      "type": "string",
      "doc": "Definition of business node",
      "Searchable": {}
    },
    {
      "name": "parentNode",
      "type": [
        "null",
        "string"
      ],
      "doc": "Parent node of the glossary term",
      "default": null,
      "Relationship": {
        "entityTypes": [
          "glossaryNode"
        ],
        "name": "IsPartOf"
      },
      "Searchable": {
        "fieldName": "parentNode",
        "fieldType": "URN",
        "hasValuesFieldName": "hasParentNode"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.GlossaryNodeUrn"
      }
    },
    {
      "name": "name",
      "type": [
        "null",
        "string"
      ],
      "doc": "Display name of the node",
      "default": null,
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldName": "displayName",
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "glossaryNodeInfo"
  }
}